import datetime
import glob
import os
import re
import logging
from typing import Dict, List, Optional, Tuple, Union
import mailbox
from mailbox import mboxMessage
from pathlib import Path
import numpy as np
import pandas as pd
from config.config import CONFIG

from bigbang.analysis import utils
from bigbang.data_types import Message, MailList, MailListDomain

filepath_auth = CONFIG.config_path + "authentication.yaml"
directory_project = str(Path(os.path.abspath(__file__)).parent.parent)
logging.basicConfig(
    filename=directory_project + "/listserv.log",
    filemode="w",
    level=logging.INFO,
    format="%(asctime)s %(message)s",
)
logger = logging.getLogger(__name__)


def email_to_dict(msg: Message) -> Dict[str, str]:
    """
    Handles data type transformation from `mailbox.mboxMessage` to `Dictionary`.
    """
    dic = {"Body": msg.get_payload()}
    for key, value in msg.items():
        dic[key] = value
    return dic


def email_to_pandas_dataframe(msg: Message) -> pd.DataFrame:
    """
    Handles data type transformation from `mailbox.mboxMessage` to `pandas.DataFrame`.
    """
    return pd.DataFrame(
        email_to_dict(msg),
        index=[msg["Message-ID"]],
    )


def email_to_mbox(msg: Message, filepath: str, mode: str = "w") -> None:
    """
    Saves `mailbox.mboxMessage` as .mbox file.
    """
    if Path(filepath).is_file():
        Path(filepath).unlink()
    mbox = mailbox.mbox(filepath)
    mbox.lock()
    mbox.add(msg)
    mbox.flush()
    mbox.unlock()


def mlist_from_mbox_to_pandas_dataframe(filepath: str) -> pd.DataFrame:
    """
    Reads `mailbox.mboxMessage` objects from .mbox file and transforms it to a
    `pandas.DataFrame`.
    For a clearer definition on what a mailing list is, see:
    bigbang.ingress.abstract.AbstractList
    """
    box = mailbox.mbox(filepath, create=False)
    mlist = []
    for msg in list(box.values()):
        _msg = dict(msg)
        if msg.is_multipart():
            # if message has many submessages, only get the top one
            # TODO: it does not detect that
            # https://list.etsi.org/scripts/wa.exe?A2=3GPP_TSG_SA_WG3_LI;6b0e2163.2008B&S=
            # is multipart
            payloads = msg.get_payload()
            _msg["body"] = str(payloads[0].get_payload(decode=True))
        else:
            _msg["body"] = str(msg.get_payload(decode=True))
        mlist.append(_msg)
    mlist = [{k.lower(): v for k, v in msg.items()} for msg in mlist]
    if len(mlist) > 0:
        # TODO: find out why some fields are missing in some messages
        df = pd.DataFrame(mlist)
        # df = utils.clean_addresses(df)
        df = utils.clean_subject(df)
        # df = utils.clean_datetime(df)
        return df
    else:
        return None


def mlist_from_mbox(filepath: str) -> list:
    """
    Reads `mailbox.mboxMessage` objects from .mbox file.
    For a clearer definition on what a mailing list is, see:
    bigbang.ingress.abstract.AbstractList
    """
    box = mailbox.mbox(filepath, create=False)
    return list(box.values())


def mlist_to_dict(
    msgs: MailList,
    include_body: bool = True,
) -> Dict[str, List[str]]:
    """
    Handles data type transformation from a List[mailbox.mboxMessage] to a
    Dictionary.
    For a clearer definition on what a mailing list is, see:
    bigbang.ingress.abstract.AbstractList
    """
    dic = {}
    for idx, msg in enumerate(msgs):
        # if msg["message-id"] != None:  #TODO: why are some 'None'?
        for key, value in msg.items():
            key = key.lower()
            if key not in dic.keys():
                dic[key] = [np.nan] * len(msgs)
            dic[key][idx] = value
    if include_body:
        dic["body"] = [np.nan] * len(msgs)
        for idx, msg in enumerate(msgs):
            # if msg["message-id"] != None:
            dic["body"][idx] = msg.get_payload()
    lengths = [len(value) for value in dic.values()]
    assert all([diff == 0 for diff in np.diff(lengths)])
    return dic


def mlist_to_pandas_dataframe(
    msgs: MailList,
    include_body: bool = True,
) -> pd.DataFrame:
    """
    Handles data type transformation from a List[mailbox.mboxMessage] to a
    pandas.DataFrame.
    For a clearer definition on what a mailing list is, see:
    bigbang.ingress.abstract.AbstractList
    """
    dic = mlist_to_dict(msgs, include_body)
    df = pd.DataFrame(dic)
    # filter out messages with unrecognisable datetime
    index = np.array(
        [
            True
            if (isinstance(dt, str)) and (len(dt) > 10) and not pd.isna(dt)
            else False
            for i, dt in enumerate(df["date"])
        ],
        dtype="bool",
    )
    df = df.loc[index, :]
    # convert data type from string to datetime.datetime object
    df.loc[:, "date"].update(
        df.loc[:, "date"].apply(
            lambda x: datetime.datetime.strptime(x, "%a, %d %b %Y %H:%M:%S %z")
        )
    )
    return df


def mlist_to_mbox(msgs: MailList, dir_out: str, filename: str) -> None:
    """
    Saves a List[mailbox.mboxMessage] as .mbox file.
    For a clearer definition on what a mailing list is, see:
    bigbang.ingress.abstract.AbstractList
    """
    # create directory path if it doesn't exist yet
    Path(dir_out).mkdir(parents=True, exist_ok=True)
    # create filepath
    filepath = f"{dir_out}/{filename}.mbox"
    # delete file if there is one at the filepath
    if Path(filepath).is_file():
        Path(filepath).unlink()
    mbox = mailbox.mbox(filepath)
    mbox.lock()
    for msg in msgs:
        try:
            mbox.add(msg)
        except Exception as e:
            logger.info(
                f'Add to .mbox error for {msg["Archived-At"]} because, {e}'
            )
    mbox.flush()
    mbox.unlock()
    logger.info(f"The list {filename} is saved at {filepath}.")
    mbox.lock()
    mbox.add(msg)
    mbox.flush()
    mbox.unlock()


def mlistdom_to_dict(
    mlists: MailListDomain, include_body: bool = True
) -> Dict[str, List[str]]:
    """
    Handles data type transformation from a List[AbstractList] to a
    `Dictionary`.
    For a clearer definition on what a mailing archive is, see:
    bigbang.ingress.abstract.AbstractArchive
    """
    nr_msgs = 0
    for ii, mlist in enumerate(mlists):
        dic_mlist = mlist_to_dict(mlist.messages, include_body)
        if ii == 0:
            dic_mlistdom = dic_mlist
            dic_mlistdom["mailing-list"] = [mlist.name] * len(mlist)
        else:
            # add mlist items to mlistdom
            for key, value in dic_mlist.items():
                if key not in dic_mlistdom.keys():
                    dic_mlistdom[key] = [np.nan] * nr_msgs
                dic_mlistdom[key].extend(value)
            # if mlist does not contain items that are in mlistdom
            key_miss = list(set(dic_mlistdom.keys()) - set(dic_mlist.keys()))
            key_miss.remove("mailing-list")
            for key in key_miss:
                dic_mlistdom[key].extend([np.nan] * len(mlist))

            dic_mlistdom["mailing-list"].extend([mlist.name] * len(mlist))
        nr_msgs += len(mlist)
    lengths = [len(value) for value in dic_mlistdom.values()]
    assert all([diff == 0 for diff in np.diff(lengths)])
    return dic_mlistdom


def mlistdom_to_pandas_dataframe(
    mlists: MailListDomain, include_body: bool = True
) -> pd.DataFrame:
    """
    Handles data type transformation from a List[AbstractList] to a
    pandas.DataFrame.
    For a clearer definition on what a mailing archive is, see:
    bigbang.ingress.abstract.AbstractArchive
    """
    df = pd.DataFrame(mlistdom_to_dict(mlists, include_body)).set_index(
        "message-id"
    )
    # get index of date-times
    index = np.array(
        [
            True
            if (isinstance(dt, str)) and (len(dt) > 10) and not pd.isna(dt)
            else False
            for i, dt in enumerate(df["date"])
        ],
        dtype="bool",
    )
    # convert data type from string to datetime.datetime object
    df.loc[index, "date"].update(
        df.loc[index, "date"].apply(
            lambda x: datetime.datetime.strptime(x, "%a, %d %b %Y %H:%M:%S %z")
        )
    )
    return df


def mlistdom_to_mbox(mlists: MailListDomain, dir_out: str):
    """
    Saves a List[AbstractList] as .mbox file.
    For a clearer definition on what a mailing archive is, see:
    bigbang.ingress.abstract.AbstractArchive
    """
    for mlist in mlists:
        mlist.to_mbox(dir_out)


def get_paths_to_files_in_directory(
    directory: str, file_dsc: str = "*"
) -> List[str]:
    """Get paths of all files matching file_dsc in directory"""
    template = f"{directory}{file_dsc}"
    file_paths = glob.glob(template)
    return file_paths


def get_paths_to_dirs_in_directory(
    directory: str, folder_dsc: str = "*"
) -> List[str]:
    """Get paths of all directories matching file_dsc in directory"""
    template = f"{directory}{folder_dsc}"
    dir_paths = glob.glob(template)
    # normalize directory paths
    dir_paths = [dir_path + "/" for dir_path in dir_paths]
    return dir_paths



"""
Miscellaneous utility functions used in other modules.
"""
import os
import time
import logging
import glob
import yaml
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from bs4 import BeautifulSoup
import requests

import networkx as nx
import pandas as pd

from config.config import CONFIG

filepath_auth = CONFIG.config_path + "authentication.yaml"
directory_project = str(Path(os.path.abspath(__file__)).parent.parent)
logging.basicConfig(
    filename=directory_project + "/scraping.log",
    filemode="w",
    level=logging.INFO,
    format="%(asctime)s %(message)s",
)
logger = logging.getLogger(__name__)


def get_paths_to_files_in_directory(
    directory: str, file_dsc: str = "*"
) -> List[str]:
    """Get paths of all files matching file_dsc in directory"""
    template = f"{directory}{file_dsc}"
    file_paths = glob.glob(template)
    return file_paths


def get_paths_to_dirs_in_directory(
    directory: str, folder_dsc: str = "*"
) -> List[str]:
    """Get paths of all directories matching file_dsc in directory"""
    template = f"{directory}{folder_dsc}"
    dir_paths = glob.glob(template)
    # normalize directory paths
    dir_paths = [dir_path + "/" for dir_path in dir_paths]
    return dir_paths


def labeled_blockmodel(g, partition):
    """
    Perform blockmodel transformation on graph *g*
    and partition represented by dictionary *partition*.
    Values of *partition* are used to partition the graph.
    Keys of *partition* are used to label the nodes of the
    new graph.
    """
    new_g = nx.quotient_graph(g, list(partition.values()), relabel=True)
    labels = dict(enumerate(partition.keys()))
    new_g = nx.relabel_nodes(new_g, labels)

    return new_g


def repartition_dataframe(df, partition):
    """
    Create a new dataframe with the same index as
    argument dataframe *df*, where columns are
    the keys of dictionary *partition*.
    The data of the returned dataframe are the
    combinations of columns listed in the keys of
    *partition*
    """
    df2 = pd.DataFrame(index=df.index)

    for k, v in list(partition.items()):
        df2[k] = df[v[0]]

        for i in range(len(v) - 1):
            df2[k] = df2[k] + df[v[i]]

    return df2


def get_common_head(str1, str2, delimiter=None):
    try:
        if str1 is None or str2 is None:
            return "", 0
        else:
            # this is ugly control flow clean it
            if delimiter is not None:

                dstr1 = str1.split(delimiter)
                dstr2 = str2.split(delimiter)

                for i in range(len(dstr1)):
                    if dstr1[:i] != dstr2[:i]:
                        # print list1[:i], list2[:i]
                        return delimiter.join(dstr1[: i - 1]), i - 1
                return str1, i
            else:
                for i in range(len(str1)):
                    if str1[:i] != str2[:i]:
                        # print list1[:i], list2[:i]
                        return str1[: i - 1], i - 1
                return str1[:i], i

        return "", 0
    except Exception as e:
        print(e)
        return "", 0


def get_common_foot(str1, str2, delimiter=None):
    head, ln = get_common_head(str1[::-1], str2[::-1], delimiter=delimiter)

    return head[::-1], ln


# turn these into automated tests
# print get_common_head('abcdefghijklmnop','abcde12345')
# print get_common_head('abcdefghijklmnop',None)
# print get_common_head('abcde\nfghijk\nlmnop\nqrst','abcde\nfghijk\nlmnopqr\nst',delimiter='\n')


def remove_quoted(mess):
    message = [
        line
        for line in mess.split("\n")
        if len(line) != 0 and line[0] != ">" and line[-6:] != "wrote:"
    ]
    new = "\n".join(message)
    return new


## remove this when clean_message is added to generic libraries
def clean_message(mess):
    if mess is None:
        mess = ""
    mess = remove_quoted(mess)
    mess = mess.strip()
    return mess


# From here:
# https://stackoverflow.com/questions/46217529/pandas-datetimeindex-frequency-is-none-and-cant-be-set
def add_freq(idx, freq=None):
    """Add a frequency attribute to idx, through inference or directly.

    Returns a copy.  If `freq` is None, it is inferred.
    """

    idx = idx.copy()
    if freq is None:
        if idx.freq is None:
            freq = pd.infer_freq(idx)
        else:
            return idx
    idx.freq = pd.tseries.frequencies.to_offset(freq)
    if idx.freq is None:
        raise AttributeError(
            "no discernible frequency found to `idx`.  Specify"
            " a frequency string with `freq`."
        )
    return idx



import sys
import subprocess
from setuptools import setup
from setuptools.command.develop import develop


class PreCommitCommand(develop):
    """
    Install pre-commit hook after all required packages are installed.
    """

    def run(self):
        try:
            cmd = (sys.executable, "-m", "pre_commit", "install")
            if not subprocess.call(cmd):
                return
        except Exception as e:
            raise SystemExit(f'Unable to run "pre-commit install" because {e}')
        develop.run(self)


# required dependencies
install_requires = [
    "beautifulsoup4>=4.9.3",
    "chardet>=4.0.0",
    "common>=0.1.2",
    "enlighten>=1.7.2",
    "gender-detector",
    "GitPython",
    "html2text>=2020.1.16",
    "ietfdata>=0.4.0",
    "ipython>=7.22.0",
    "jinja2>=2.11.3",
    "jupyter>=1.0.0",
    "jsonschema>=3.2.0",
    "matplotlib>=3.4.1",
    "networkx>=2.5.1",
    "nltk>=3.6.2",
    "numpy>=1.20.2",
    "pandas>=1.2.3",
    "powerlaw>=1.4.6",
    "python-dateutil>=2.8.1",
    "python-Levenshtein>=0.12.2",
    "pytz>=2021.1",
    "pyyaml>=5.4.1",
    "pyzmq>=22.0.3",
    "requests>=2.25.1",
    "seaborn",
    "tornado>=6.1",
    "tqdm>=4.60.0",
    "validator_collection>=1.5.0",
    "python-docx>=0.8",
]
# test dependencies
test_requires = ["pytest>=6.2.3", "coverage>=5.5", "testfixtures>=6.17.1"]
# development dependencies
dev_requires = [
    "black>=20.8b1",
    "isort>=5.7.0",
    "pre-commit>=2.13.0",
]

config = {
    "name": "BigBang",
    "version": "0.4.1",
    "description": "BigBang is a toolkit for studying communications data from collaborative projects. It currently supports analyzing mailing lists from Sourceforge, Mailman, ListServ, or .mbox files.",
    "author": "BigBang Team",
    "author_email": "bigbang-dev@data-activism.net",
    "packages": ["bigbang"],
    "install_requires": install_requires,
    "extras_require": {
        "test": test_requires,
        "dev": test_requires + dev_requires,
    },
    "cmdclass": {"develop": PreCommitCommand},
}

setup(**config)



"""
This module supports the Archive class, a generic
structure representing a collection of archived emails,
typically from a single mailing list.
"""

import datetime
import logging
import mailbox

import numpy as np
import os
import pandas as pd
from pathlib import Path
import pytz

import bigbang.analysis.utils as analysis_utils
import bigbang.ingress.mailman as mailman
from bigbang.parse import get_date, get_text
import bigbang.analysis.process as process
from bigbang.analysis.thread import Node, Thread
from config.config import CONFIG

from . import utils


def load(path):
    data = pd.read_csv(path)
    return Archive(data)


class ArchiveWarning(BaseException):
    """Base class for Archive class specific exceptions"""

    pass


class MissingDataException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Archive(object):
    """A representation of a mailing list archive."""

    data = None
    activity = None
    threads = None
    entities = None
    preprocessed = None

    def __init__(self, data, archive_dir=CONFIG.mail_path, mbox=False):
        """
        Initialize an Archive object.

        The behavior of the constructor depends on the type
        of its first argument, data.

        If data is a Pandas DataFrame, it is treated as a representation of
        email messages with columns for Message-ID, From, Date, In-Reply-To,
        References, and Body. The created Archive becomes a wrapper around a
        copy of the input DataFrame.

        If data is a string, then it is interpreted as a path to either a
        single .mbox file (if the optional argument single_file is True) or
        else to a directory of .mbox files (also in .mbox format). Note that
        the file extensions need not be .mbox; frequently they will be .txt.

        Upon initialization, the Archive object drops duplicate entries
        and sorts its member variable *data* by Date.

        Parameters
        ----------
        data : pandas.DataFrame, or str
        archive_dir: str, optional
            Defaults to CONFIG.mail_path
        mbox: bool

        """

        if isinstance(data, pd.core.frame.DataFrame):
            self.data = data.copy()
        elif isinstance(data, str):
            self.data = load_data(data, archive_dir=archive_dir, mbox=mbox)

        try:
            self.data["Date"] = pd.to_datetime(
                self.data["Date"],
                errors="coerce",
                infer_datetime_format=True,
                utc=True,
            )
        except Exception as e:
            logging.error(
                "Error while converting to datetime, despite coerce mode. "
                + f"The error message is: {e}"
            )
            raise ArchiveWarning(
                "Error while converting to datetime, despite coerce mode. "
                + f"The error message is: {e}"
            )

        try:
            self.data.drop_duplicates(inplace=True)
        except Exception as e:
            logging.error(
                "Error while removing duplicate messages, maybe timezone issues?"
                + f"The error message is: {e}",
                exc_info=True,
            )

        # Drops any entries with no Date field.
        # It may be wiser to optionally
        # do interpolation here.
        if self.data["Date"].isnull().any():
            # self.data.dropna(subset=['Date'], inplace=True)
            self.data = self.data[self.data["Date"].notnull()]
            # workaround for https://github.com/pandas-dev/pandas/issues/13407

        # convert any null fields to None -- csv saves these as nan sometimes
        self.data = self.data.where(pd.notnull(self.data), None)
        self.preprocessed = {}

        try:
            # set the index to be the Message-ID column
            self.data.set_index("Message-ID", inplace=True)
        except KeyError:
            # will get KeyError if Message-ID is already index
            pass

        # let's do a pass to try to find bad tzinfo's
        bad_indices = []
        for i, row in self.data.iterrows():
            try:
                row["Date"] < datetime.datetime.now(pytz.utc)
            except Exception as e:
                logging.error(
                    "Error timezone issues while detecting bad rows. "
                    + f"The error message is: {e}",
                    exc_info=True,
                )
                bad_indices.append(i)
                logging.info("Bad timezone on %s", row["Date"])
        if len(bad_indices) > 0:
            # drop those rows that threw an error
            self.data = self.data.drop(bad_indices)
            logging.info("Dropped %d rows", len(bad_indices))

        try:
            self.data.sort_values(by="Date", inplace=True)
        except Exception as e:
            logging.error(
                "Error while sorting, maybe timezone issues?"
                + f" The error message is: {e}",
                exc_info=True,
            )

        if self.data.empty:
            raise MissingDataException(
                "Archive after initial processing is empty. Was data collected properly?"
            )

    def add_affiliation(self, rel_email_affil):
        """
        Uses a DataFrame of email affiliation information and adds it to the archive's data table.

        The email affilation data is expected to have a regular format, with columns:

          - ``email`` - strings, complete email addresses
          - ``affilation`` - strings, names of organizatiosn of affilation
          - ``min_date`` - datetime, the starting date of the affiliation
          - ``max_date`` - datetime,the end date of the affilation.

        Note that this mutates the dataframe in ``self.data`` to add the affiliation data.

        Params
        ------

        rel_email_affil : pandas.DataFrame
        """

        def match_affiliation(row):
            ## TODO: Option to turn off the max or min constraints in case of limited data.
            matches = rel_email_affil[
                (
                    rel_email_affil["email"]
                    == analysis_utils.extract_email(row["From"])
                )
                & (
                    rel_email_affil["min_date"].dt.tz_localize("utc")
                    <= pd.to_datetime(row["Date"])
                )
                & (
                    rel_email_affil["max_date"].dt.tz_localize("utc")
                    >= pd.to_datetime(row["Date"])
                )
            ]
            return (
                matches["affiliation"].values[0]
                if matches.shape[0] > 0
                else None
            )

        self.data["affiliation"] = self.data.apply(match_affiliation, axis=1)

    def resolve_entities(self, inplace=True):
        """Return data with resolved entities.

        Parameters
        ----------
        inplace : bool, default True

        Returns
        -------
        pandas.DataFrame or None
            Returns None if inplace == True

        """
        if self.entities is None:
            if self.activity is None:
                self.get_activity()

            self.entities = process.resolve_sender_entities(self.activity)

        to_replace = []

        value = []

        for e, names in list(self.entities.items()):
            for n in names:
                to_replace.append(n)
                value.append(e)

        data = self.data.replace(
            to_replace=to_replace, value=value, inplace=inplace
        )

        # clear and replace activity with resolved activity
        self.activity = None
        self.get_activity()

        if inplace:
            return self.data
        else:
            return data

    def get_activity(self, resolved=False):
        """
        Get the activity matrix of an Archive.

        Columns of the returned DataFrame are the Senders of emails.
        Rows are indexed by ordinal date.
        Cells are the number of emails sent by each sender on each data.

        If *resolved* is true, then default entity resolution is run on the
        activity matrix before it is returned.
        """
        if self.activity is None:
            self.activity = self.compute_activity(self)

        if resolved:
            self.entities = process.resolve_sender_entities(self.activity)
            eact = utils.repartition_dataframe(self.activity, self.entities)

            return eact

        return self.activity

    def compute_activity(self, clean=True):
        """Return the computed activity."""
        mdf = self.data

        if clean:
            # unnecessary?
            if mdf["Date"].isnull().any():
                mdf = mdf.dropna(subset=["Date"])

            mdf = mdf[
                mdf["Date"] < datetime.datetime.now(pytz.utc)
            ]  # drop messages apparently in the future

        mdf2 = mdf.reindex(columns=["From", "Date"])
        mdf2["Date"] = mdf["Date"].apply(lambda x: x.toordinal())

        activity = (
            mdf2.groupby(["From", "Date"]).size().unstack("From").fillna(0)
        )
        new_date_range = np.arange(mdf2["Date"].min(), mdf2["Date"].max())
        # activity.set_index('Date')
        activity = activity.reindex(new_date_range, fill_value=0)

        return activity

    def get_personal_headers(self, header="From"):
        """
        Returns a dataframe with a row for every message of the archive, containing
        column entries for:

         - The personal header specified. Defaults to "From". Could be "Repy-To".
         - The email address extracted from the From field
         - The domain of the From field

        This dataframe is computed the first time this method is called and then cached.


        Parameters
        ------------

        header: string, default "From"

        Returns
        ----------

        data: pandas.DataFrame
        """
        if header in self.preprocessed:
            return self.preprocessed[header]
        else:
            emails = self.data[header].apply(analysis_utils.extract_email)
            domains = self.data[header].apply(analysis_utils.extract_domain)
            self.preprocessed[header] = pd.concat(
                [self.data[header], emails, domains],
                axis=1,
                keys=[header, "email", "domain"],
            )
            return self.preprocessed[header]

    def get_threads(self, verbose=False):
        """Get threads."""

        if self.threads is not None:
            return self.threads

        df = self.data

        threads = list()
        visited = dict()

        total = df.shape[0]
        c = 0

        for i in df.iterrows():

            if verbose:
                c += 1
                if c % 1000 == 0:
                    logging.info("Processed %d of %d" % (c, total))

            if i[1]["In-Reply-To"] == "None":
                root = Node(i[0], i[1])
                visited[i[0]] = root
                threads.append(Thread(root))
            elif i[1]["In-Reply-To"] not in list(visited.keys()):
                root = Node(i[1]["In-Reply-To"])
                succ = Node(i[0], i[1], root)
                root.add_successor(succ)
                visited[i[1]["In-Reply-To"]] = root
                visited[i[0]] = succ
                threads.append(Thread(root, known_root=False))
            else:
                parent = visited[i[1]["In-Reply-To"]]
                node = Node(i[0], i[1], parent)
                parent.add_successor(node)
                visited[i[0]] = node

        self.threads = threads

        return threads

    def save(self, path, encoding="utf-8"):
        """Save data to csv file."""
        self.data.to_csv(path, ",", encoding=encoding)


def archive_directory(base_dir, list_name):
    """
    Creates a new archive directory for the given list_name unless one already exists.
    Returns the path of the archive directory.
    """
    arc_dir = os.path.join(base_dir, list_name)
    Path(arc_dir).mkdir(parents=True, exist_ok=True)
    return arc_dir


def find_footer(messages, number=1):
    """
    Returns the footer of a DataFrame of emails.

    A footer is a string occurring at the tail of most messages.
    Messages can be a DataFrame or a Series
    """
    if isinstance(messages, pd.DataFrame):
        messages = messages["Body"]

    # sort in lexical order of reverse strings to maximize foot length
    srb = messages.apply(lambda x: None if x is None else x[::-1]).order()
    # srb = df.apply(lambda x: None if x['Body'] is None else x['Body'][::-1],
    #              axis=1).order()
    # begin walking down the series looking for maximal overlap
    counts = {}

    last = None

    def clean_footer(foot):
        return foot.strip()

    for b in srb:
        if last is None:
            last = b
            continue
        elif b is None:
            continue
        else:
            head, i = utils.get_common_head(b, last, delimiter="\n")
            head = clean_footer(head[::-1])
            last = b

            if head in counts:
                counts[head] = counts[head] + 1
            else:
                counts[head] = 1

        last = b

    # reduce candidates that are strictly longer and less frequent
    # than most promising footer candidates
    for n, foot1 in sorted(
        [(v, k) for k, v in list(counts.items())], reverse=True
    ):
        for foot2, m in list(counts.items()):
            if n > m and foot1 in foot2 and len(foot1) > 0:
                counts[foot1] = counts[foot1] + counts[foot2]
                del counts[foot2]

    candidates = sorted(
        [(v, k) for k, v in list(counts.items())], reverse=True
    )

    return candidates[0:number]


def load_data(
    name: str, archive_dir: str = CONFIG.mail_path, mbox: bool = False
):
    """
    Load the data associated with an archive name, given as a string.

    Attempt to open {archives-directory}/NAME.csv as data.

    Failing that, if the the name is a URL, it will try to derive
    the list name from that URL and load the .csv again.

    Parameters
    ----------
    name : str
    archive_dir : str, default CONFIG.mail_path
    mbox : bool, default False
        If true, expects and opens an mbox file at this path

    Returns
    -------
    data : pandas.DataFrame

    """

    if mbox:
        return open_list_archives(name, archive_dir=archive_dir, mbox=True)

    # a first pass at detecting if the string is a URL...
    if not (name.startswith("http://") or name.startswith("https://")):
        path = os.path.join(archive_dir, name + ".csv")

        if os.path.exists(path):
            data = pd.read_csv(path)
            return data
        else:
            logging.warning("No data available at %s", path)
    else:
        # TODO: Should "get_list_name" be more general than mailman?
        path = os.path.join(archive_dir, mailman.get_list_name(name) + ".csv")

        if os.path.exists(path):
            data = pd.read_csv(path)
            return data
        else:
            logging.warning(
                "No data found for %s. Check directory name and whether archives have been collected.",
                name,
            )


def messages_to_dataframe(messages):
    """
    Turn a list of parsed messages into a dataframe of message data,
    indexed by message-id, with column-names from headers.
    """
    # extract data into a list of tuples -- records -- with
    # the Message-ID separated out as an index
    # valid_messages = [m for m in messages if m.get()

    pm = [
        (
            m.get("Message-ID"),
            str(m.get("From")).replace("\\", " "),
            str(m.get("Subject")),
            get_date(m),
            str(m.get("In-Reply-To")),
            str(m.get("References")),
            get_text(m),
        )
        for m in messages
        if m.get("From")
    ]

    mdf = pd.DataFrame.from_records(
        list(pm),
        index="Message-ID",
        columns=[
            "Message-ID",
            "From",
            "Subject",
            "Date",
            "In-Reply-To",
            "References",
            "Body",
        ],
    )
    mdf.index.name = "Message-ID"

    return mdf


def open_list_archives(
    archive_name: str,
    archive_dir: str = CONFIG.mail_path,
    mbox: bool = False,
) -> pd.DataFrame:
    """
    Return a list of all email messages contained in the specified directory.

    Parameters
    -----------
    archive_name: str
        the name of a subdirectory of the directory specified
        in argument *archive_dir*. This directory is expected to contain
        files with extensions .txt, .mail, or .mbox. These files are all
        expected to be in mbox format-- i.e. a series of blocks of text
        starting with headers (colon-separated key-value pairs) followed by
        an email body.

    archive_dir : str:
        directory containing all messages.

    mbox: bool, default False
        True if there's an mbox file already available for this archive.

    Returns
    --------

    data : pandas.DataFrame

    """
    messages = None

    if mbox and (os.path.isfile(os.path.join(archive_dir, archive_name))):
        # treat string as the path to a file that is an mbox
        box = mailbox.mbox(
            os.path.join(archive_dir, archive_name), create=False
        )
        messages = list(box.values())

    else:
        # assume string is the path to a directory with many
        # TODO: Generalize get_list_name to listserv, w3c?
        list_name = mailman.get_list_name(archive_name)
        arc_dir = archive_directory(archive_dir, list_name)

        file_extensions = [".txt", ".mail", ".mbox"]

        txts = [
            os.path.join(arc_dir, fn)
            for fn in os.listdir(arc_dir)
            if any([fn.endswith(extension) for extension in file_extensions])
        ]

        logging.info(os.listdir(arc_dir))
        logging.info("Opening %d archive files", len(txts))

        def mbox_reader(stream):
            """Read a non-ascii message from mailbox"""
            data = stream.read()
            text = data.decode(encoding="utf-8", errors="replace")
            return mailbox.mboxMessage(text)

        arch = [
            mailbox.mbox(txt, factory=mbox_reader, create=False).values()
            for txt in txts
        ]

        if len(arch) == 0:
            raise MissingDataException(
                (
                    "No messages in %s under %s. Did you run the "
                    "collect_mail.py script?"
                )
                % (archive_dir, list_name)
            )

        messages = [item for sublist in arch for item in sublist]

    return messages_to_dataframe(messages)
